/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2006 bonbons <bonbons67@internet.lu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include "../../src/initng.h"

#include <assert.h>
#include <stdio.h>
#include <mntent.h>
#include <stdlib.h>                         /* free() exit() */
#include <string.h>
#include <sys/stat.h>
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"

struct fstab_entry
{
    char *mountpoint;
    char *device;
    char *options;
    char *fstype;
    int freq;
    int pass;
    struct fstab_entry *next;
};

static void free_fstab(struct fstab_entry *tab);
static struct fstab_entry *load_fstab(const char *file);
static void check_tabs(void);
static int check_START_DEP_MET(active_db_h * service);

/* fstab information */
struct fstab_entry *fstab = NULL;
time_t fstab_date = 0;
off_t fstab_size = 0;

/* /proc/mounts information */
struct fstab_entry *mounts = NULL;
time_t mounts_date = 0;

s_entry REQUIRE_MOUNT = { "require_mount", STRINGS, NULL,
    "If set, initng will wait for listed mountpoints to be mounted before starting this service."
};

/**
 * Free memory used by our fstab copy
 */
static void free_fstab(struct fstab_entry *tab)
{
    while (tab)
    {
        struct fstab_entry *next = tab->next;

        free(tab->mountpoint);
        free(tab->device);
        free(tab->options);
        free(tab->fstype);
        free(tab);
        tab = next;
    }
}

/**
 * Check if fstab exists, if it was modified since last check
 * and relod if our cached information is non-existent/out of date
 */
static struct fstab_entry *load_fstab(const char *file)
{
    FILE *f;

    if ((f = setmntent(file ? file : "/etc/fstab", "r")))
    {
        struct mntent *line;
        struct fstab_entry *last = NULL;
        struct fstab_entry *first = NULL;
        int n = 0;

        while ((line = getmntent(f)))
        {
            struct fstab_entry *tmp = malloc(sizeof(struct fstab_entry));

            if (tmp == NULL)
                continue;
            tmp->mountpoint = strdup(line->mnt_dir);
            tmp->device = strdup(line->mnt_fsname);
            tmp->options = strdup(line->mnt_opts);
            tmp->fstype = strdup(line->mnt_type);
            tmp->freq = line->mnt_freq;
            tmp->pass = line->mnt_passno;
            tmp->next = NULL;
            if (first && last)
                last->next = tmp;
            else
            {
                last = tmp;
                first = tmp;
            }
            n++;
        }
        endmntent(f);
        return first;
    }
    return NULL;
}

static void check_tabs(void)
{
    struct stat st;
    struct fstab_entry *tmp;

    /* check /etc/fstab */
    if (stat("/etc/fstab", &st) == 0)
    {
        if (st.st_mtime == fstab_date && st.st_size == fstab_size
            && (fstab || st.st_size == 0))
            goto mounts;

        P_("/etc/fstab has changed, reloading!");
        tmp = fstab;
        fstab = NULL;
        fstab_date = st.st_mtime;
        fstab_size = st.st_size;
        free_fstab(tmp);
        fstab = load_fstab("/etc/fstab");
    }
    else
        W_("Could not stat /etc/fstab");

  mounts:
    /* check /proc/mounts */
    if (mounts_date != g.now.tv_sec)
        mounts = load_fstab("/proc/mounts");
    mounts_date = g.now.tv_sec;
}

static int check_mountpoint(const char *path)
{
    struct stat st;
    const struct fstab_entry *lfstab = fstab;

    /* We don't like these unset or relative paths */
    if (path == NULL || *path != '/')
        return FAIL;

    /* We always assume / as being mounted */
    if (strcmp(path, "/") == 0)
        return TRUE;

    /* If we can't stat the path something is missing anyhow */
    if (stat(path, &st))
        return FALSE;

    /* We only expect checking for mountpoint, so it must be a dir!
     * If checking for files, use plugin fstat */
    if (!S_ISDIR(st.st_mode))
        return FAIL;

    /* Now check path against fstab, and verify any matching against mounts */
    while (lfstab)
    {
        int l = strlen(lfstab->mountpoint);

        if (strncmp(lfstab->mountpoint, path, l) == 0)
        {
            /* Mointpoint is on path, lookup in mounts */
            const struct fstab_entry *mnt = mounts;

            while (mnt && strcmp(mnt->mountpoint, lfstab->mountpoint))
                mnt = mnt->next;

            if (mnt == NULL)
                return FALSE;               /* Not mounted */
        }
        lfstab = lfstab->next;
    }

    /* If we get here at least one of the following conditions is met:
     * - path does not contain a mountpoint from fstab
     * - all mountpoints in path exist in mounts
     */
    return TRUE;
}

static int check_START_DEP_MET(active_db_h * service)
{
    const char *apath = NULL;

    assert(service);

    /* only apply if service requires mountpoints */
    if (!initng_active_db_is(&REQUIRE_MOUNT, service))
        return (TRUE);

    D_("Doing check because REQUIRE_MOUNT is set.\n");

    check_tabs();

    while ((apath =
            initng_active_db_get_next_string(&REQUIRE_MOUNT, service, apath)))
    {
        int r = check_mountpoint(apath);

        if (r == FAIL)
        {
            F_("Failed to check for mountpoint '%s'", apath);
            return FAIL;
        }
        else if (r == FALSE)
        {
            D_("Mount point '%s' not ready for %s", apath, service->name);
            goto miss;
        }
    }
    return TRUE;

  miss:
    /* Make sure mainloop will run within 1 second. */
    initng_global_set_sleep(1);

    return (FALSE);
}

int module_init(int api_version)
{
    S_;
    if (api_version != API_VERSION)
    {
        F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", INITNG_VERSION, api_version);
        return (FALSE);
    }

    initng_service_data_types_add(&REQUIRE_MOUNT);
    initng_plugin_hook_add(&g.START_DEP_MET, 55, &check_START_DEP_MET);
    return (TRUE);
}

void module_unload(void)
{
    S_;
    initng_service_data_types_del(&REQUIRE_MOUNT);
    initng_plugin_hook_del(&g.START_DEP_MET, &check_START_DEP_MET);
}
